/*
 * realreplay lib
 *
 * see http://sed.free.fr/realplay for the documentation.
 * This is public domain.
 *
 * Emulate mode done by David Decotigny <david.decotigny@free.fr>
 * Choice of directory by David Decotigny (with adaptations)
 */

#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/select.h>

int (*the_open)(const char *, int, mode_t) = 0;
int (*the_ioctl)(int, int, ...);
int (*the_close)(int) = 0;
int (*the_write)(int, const void*, int) = 0;
pid_t (*the_fork)(void) = 0;

//int  (*the_select)(int, fd_set *, fd_set *, fd_set *, struct timeval *) = 0;

/* some local declarations */
static void emergency_exit(int);
static int stopper_watchdog(int wait_fd, int timeout_ms);

/* our static globals */
struct trace {
  void *address;
  size_t address_size;
  int size;
  int max_size;
  void **bt;
};

static unsigned int lib_init=0;
static unsigned int audio_out=0;
static unsigned int audio_emulate=0;

static char out_dirpath[FILENAME_MAX+1] = "/tmp";

static int          watchdog_pid=-1;
static int          watchdog_fd=-1;

static void init_lib(void)
{
  char *a;
  char *libc_path;
  void *p;
  int pipes[2];
  int timeout_s;

  /* useless check ? */
  if (lib_init)
    return;
  lib_init = 1;

  libc_path = getenv("LIBC_PATH");
  if (!libc_path) {
    fprintf(stderr, "error : LIBC_PATH environment variable not declared\n");
    emergency_exit(1);
  }

  /* let's interface with the libc */
  p = dlopen(libc_path, RTLD_LAZY);
  if (!p) {
    fprintf(stderr, "dlopen error. You surely gave a wrong LIBC_PATH\n");
    emergency_exit(1);
  }

  the_fork = dlsym(p, "fork");
  if (!the_fork) {
    perror("dlsym(fork) error");
    exit(1);
  }

  the_open = dlsym(p, "open");
  if (!the_open) {
    perror("dlsym(open) error");
    exit(1);
  }

  the_close = dlsym(p, "close");
  if (!the_close) {
    perror("dlsym(close) error");
    exit(1);
  }

  the_ioctl = dlsym(p, "ioctl");
  if (!the_ioctl) {
    perror("dlsym(ioctl) error");
    exit(1);
  }

  the_write = dlsym(p, "write");
  if (!the_write) {
    perror("dlsym(write) error");
    exit(1);
  }

  a = getenv("LIBREPLAY_AUDIO");
  if (a && a[0] == '1' && a[1] == 0)
    audio_out = 1;
  else {
     a = getenv("LIBREPLAY_EMULATE");
     if (a) {
       sscanf(a, "%u", &audio_emulate);
       if (audio_emulate == 0) {
         fprintf(stderr, "bad value for LIBREPLAY_EMULATE (try 1, 2, 3, ...)\n");
         exit(1);
       }
       fprintf(stderr,
	       "Emulation mode enabled, speedup is %d\n", audio_emulate);
     } else {
       fprintf(stderr, "LIBREPLAY_AUDIO not set or different from \"1\" -> fast mode\n");
       fprintf(stderr, "if you have bugs in the output, try to set this environment variable to 1\n");
       fprintf(stderr, "it will output the sound while saving it -> all is slow\n");
       fprintf(stderr, "you may also set LIBREPLAY_EMULATE to 1, 2, 3, ...\n");
       fprintf(stderr, "It will emulate output to the audio card with usleep.\n");
       fprintf(stderr, "1 is for close to realtime, 3 and above will\n");
       fprintf(stderr, "sleep at 1/3 of realtime and less (some streams like it this way)\n");
       fprintf(stderr, "(if you don't understand, contact sed@free.fr)\n");
     }
  }


  a = getenv("LIBREPLAY_DIRPATH");
  if (a) {
    /* 200 to get some room for the file's name */
    if (strlen(a) > FILENAME_MAX - 200)
      fprintf(stderr, "%s: LIBREPLAY_DIRPATH name too long, default back to /tmp\n", a);
    else {
      strncpy(out_dirpath, a, sizeof(out_dirpath));
      out_dirpath[sizeof(out_dirpath)-1] = '\0';
    }
  } else {
    fprintf(stderr, "LIBREPLAY_DIRPATH not set. If you want to write the stream to a directory\n");
    fprintf(stderr, "other than /tmp, set the variable to your wish.\n");
  }


  /* Create the FIFO and the watchdog process */
  if (watchdog_pid >= 0) {
    fprintf(stderr, "Watchdog already started: pid %d, fd %d\n",
	    watchdog_pid, watchdog_fd);
    goto after_watchdog;
  }
  a=getenv("LIBREPLAY_TIMEOUT");
  if (!a) {
    fprintf(stderr, "LIBREPLAY_TIMEOUT not set. If you want to automatically\n");
    fprintf(stderr, "kill realplayer when no sound is sent for some time, set\n");
    fprintf(stderr, "this variable to the timeout (in seconds).\n");
    goto after_watchdog;
  }
  if (sscanf(a, "%d", &timeout_s) != 1) {
    fprintf(stderr, "Warning: cannot parse watchdog timeout\n");
    goto after_watchdog;
  }
  if (!timeout_s)
    /* timeout = 0 => no timeout */
    goto after_watchdog;
  if (timeout_s < 0) {
    fprintf(stderr, "Warning: invalid timeout (%s)\n", a);
    goto after_watchdog;
  }
  if (pipe(pipes) < 0) {
    fprintf(stderr, "Cannot create pipe for watchdog\n");
    perror("");
    goto after_watchdog;
  }
  watchdog_fd = pipes[1];
  watchdog_pid = stopper_watchdog(pipes[0], timeout_s*1000);

after_watchdog: ;
}

int snd=-1;
int devdsp=-1;

static char filename[FILENAME_MAX+1];
static int streamnb = 0;
static int stream_speed=0, stream_channels=0, stream_format=0;

int close(int fd)
{
  if (!the_close)
    init_lib();
  if (!the_close) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }

  if (fd == snd) {
    snd = -1;
    fprintf(stderr, "closing stream %s\n", filename);
    if (audio_out)
      the_close(devdsp);
    devdsp = -1;
  }
  return the_close(fd);
}

int open(const char *t, int z, mode_t r)
{
  if (!the_open)
    init_lib();
  if (!the_open) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }

  fprintf(stderr, "open %s\n", t);

  if (!strcmp(t, "/dev/mixer")) {
    int fd;

    if (audio_out)
      return the_open(t, z, r);

    fd = the_open("/dev/null", O_RDWR|O_CREAT|O_TRUNC, 0644);

    fprintf(stderr, "catching up mixer (%d), opening /dev/null\n", fd);
    errno = 0;

    return fd;
  }

  if (!strcmp(t, "/dev/dsp")) {
    streamnb++;
    stream_speed = stream_channels = stream_format = 0;
    snprintf(filename, FILENAME_MAX-1,
	     "%s/realplayer_stream_%d_speed_%d_channels_%d_format_%d.raw",
	     out_dirpath, streamnb, stream_speed, stream_channels,
	     stream_format);

    if (audio_out) {
      devdsp = the_open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
      fprintf(stderr, "opening stream %s (%d %d)\n", filename, z, r);
      snd = the_open(t, z, r);
      return snd;
    }

    snd = the_open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
    fprintf(stderr, "opening stream %s (%d %d)\n", filename, z, r);
    return snd;
  }

  return the_open(t, z, r);
}

typedef struct audio_buf_info {
  int fragments; 
int fragstotal; /* Total # of fragments allocated */
int fragsize;   /* Size of a fragment in bytes */
int bytes;   /* Available space in bytes (includes partially used fragments) */
/* Note! 'bytes' could be more than fragments*fragsize */
} audio_buf_info;

static char f2[FILENAME_MAX+1];
int ioctl(int d, unsigned long request, ...)
{
  va_list ap;
  void *r;

  if (!the_ioctl)
    init_lib();
  if (!the_ioctl) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }

  va_start(ap, request);
  r = va_arg(ap, void *);
  va_end(ap);

  if (d == snd && snd != -1) {
    char c = ((unsigned)request >>8) & 0xff;
    int w = (unsigned)request & 0xff;

    if (c == 'P' && w == 0) {
      //fprintf(stderr, "reset card\n");
    } else
    if (c == 'P' && w == 2) {
      fprintf(stderr, "set speed at %d\n", *(int *)r);
      stream_speed = *(int *)r;
      snprintf(f2, FILENAME_MAX-1,
	       "%s/realplayer_stream_%d_speed_%d_channels_%d_format_%d.raw",
	       out_dirpath, streamnb, stream_speed, stream_channels,
	       stream_format);
      rename(filename, f2);
      strncpy(filename, f2, FILENAME_MAX);
    } else
    if (c == 'P' && w == 6) {
      fprintf(stderr, "set nb channels %d\n", *(int *)r);
      stream_channels = *(int *)r;
      snprintf(f2, FILENAME_MAX-1,
	       "%s/realplayer_stream_%d_speed_%d_channels_%d_format_%d.raw",
	       out_dirpath, streamnb, stream_speed, stream_channels,
	       stream_format);
      rename(filename, f2);
      strncpy(filename, f2, FILENAME_MAX);
    } else
    if (c == 'P' && w == 5) {
      fprintf(stderr, "set format %d\n", *(int *)r);
      stream_format = *(int *)r;
      snprintf(f2, FILENAME_MAX-1,
	       "%s/realplayer_stream_%d_speed_%d_channels_%d_format_%d.raw",
	       out_dirpath, streamnb, stream_speed, stream_channels,
	       stream_format);
      rename(filename, f2);
      strncpy(filename, f2, FILENAME_MAX);
    } else
    if (c == 'P' && w == 10) {
      //fprintf(stderr, "set nb fragments %8.8x\n", *(int *)r);
    } else
    if (c == 'P' && w == 12) {
      audio_buf_info *x = r;
      //fprintf(stderr, "asking for space ? ok, there we go :)\n");
      x->fragments = 10;
      x->fragstotal = 10;
      x->fragsize = 4096;
      x->bytes = 10 * 4096;
    } else
    if (c == 'P' && w == 23) {
      //fprintf(stderr, "getodelay %8.8x\n", *(int *)r);
      /* we introduce a small delay, some streams coredump realplay
       * if we don't do it, try to change this value if you have a
       * problem
       */
      *(int *)r = 10000;
    } else {
      fprintf(stderr, "unkown ioctl\n");
      fprintf(stderr, "calling ioctl on %d request=%8.8x\n", d, (unsigned)request);
      fprintf(stderr, "%c %d\n", ((unsigned)request >>8) & 0xff, (unsigned)request & 0xff);
      return -1;
    }
    if (audio_out)
      return the_ioctl(d, request, r);
    else
      return 0;
  }

  return the_ioctl(d, request, r);
}

ssize_t write(int fd, const void *buf, size_t count)
{
  if (!the_write)
    init_lib();
  if (!the_write) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }

  if (audio_out && fd == snd)
    write(devdsp, buf, count);
  else if (audio_emulate && fd == snd) {
#if 0
    unsigned bps =  stream_speed*stream_channels*stream_format;
    unsigned Bpms = bps / 8000; // Bytes per milliseconds
    unsigned ms = count / Bpms; // "RT" delay in milliseconds

    ms /= audio_emulate;

    usleep(ms * 1000);
#endif
    /* let's calculate with (hopefully) less rounding errors */
    unsigned bps = stream_speed*stream_channels*stream_format;
    unsigned ms = count * 8000 / bps;
    usleep((ms * 1000) / audio_emulate);
  }

  if (fd == snd && watchdog_fd > 0)
    the_write(watchdog_fd, "w", 1);

  return the_write(fd, buf, count);
}

static void emergency_exit(int status)
{
  kill(getpid(), SIGTERM);
}

/* we catch fork because we must init the lib before realplay forks */
pid_t fork(void)
{
  if (!the_fork)
    init_lib();
  if (!the_fork) {
    fprintf(stderr, "problem with preload stuff\n");
    return -1;
  }

  return the_fork();
}


/* Return PID of the watchdog, or -1 on error */
int stopper_watchdog(int wait_fd, int timeout_ms)
{
  int realplay_pid = getpid();
  int retval = the_fork();
  int first_run = 1;

  if (retval < 0)
    {
      fprintf(stderr, "Warning: cannot fork stopper watchdog\n");
      return retval;
    }
  if (retval != 0)
    /* parent */
    return retval;

  fprintf(stderr, "Watchdog %d (parent %d) started (timeout %dms) on FD %d.\n",
	  getpid(), realplay_pid, timeout_ms, wait_fd);

  while (1)
    {
      fd_set fds;
      struct timeval tv;
      int retval;

      char msg;

      /* Timeout before next write to /dev/dsp */
      FD_ZERO(&fds);
      FD_SET(wait_fd, &fds);
      memset(&tv, 0, sizeof(tv));
      if (!first_run) {
        tv.tv_sec  = timeout_ms / 1000;
        tv.tv_usec = (timeout_ms % 1000) * 1000;
      } else {
        /* first timeout set to 1 minute */
        tv.tv_sec  = 60;
        tv.tv_usec = 0;
        first_run = 0;
      }

      retval = select(wait_fd+1, &fds, NULL, NULL, &tv);
      if (retval < 0)
	{
	  fprintf(stderr, "Watchdog warning: select error\n");
	  break;
	}

      /* Timeout ! */
      if (! FD_ISSET(wait_fd, &fds))
	{
	  fprintf(stderr, "Watchdog timeout %d => kill realplay !\n",
		  timeout_ms);
	  kill(realplay_pid, SIGTERM);
	  break;
	}

      /* Eat message */
      retval = read(wait_fd, & msg, 1);
      if ( (retval != 1) || (msg != 'w') )
	{
	  fprintf(stderr, "Watchdog warning: invalid message (%d / %d)\n",
		  retval, msg);
	  break;
	}

    }

  exit(0);
}
